#ifndef __SHADOW_HPP__
#define __SHADOW_HPP__

#include <time.h>
#include <stdio.h>
#include <sys/types.h>

#include "config.h"
#include "entry.hpp"

#define SECONDS_PER_DAY		(24L * 60L * 60L)

// Structure definition
struct spwd 
{
	char *sp_namp;								// login name
	char *sp_pwdp;								// encrypted password
	time_t sp_lstchg;							// date of last change
	time_t sp_min;								// minimum number of days between changes
	time_t sp_max;								// maximum number of days between changes
	time_t sp_warn;							// number of days of warning before password expires
	time_t sp_inact;							// number of days after password expires until the account becomes unusable
	time_t sp_expire;							// days since 1/1/70 until account expires
	unsigned long sp_flag;					// reserved for future use
};

class CShadowEntry : public CEntry
{
	public:
		CShadowEntry(FILE *);
		CShadowEntry(char *);
		~CShadowEntry();
	
	public:
		char *GetName();
		char *GetPassword();
		char *GetGecos() { return NULL; } 			// dummy
		char *GetHome() { return NULL; } 			// dummy	
		gid_t GetGroup() { return (gid_t)-1; }		// there is no group info in shadow file
		uid_t GetUser() { return (uid_t)-1;	}		// there is no user info in shadow file
		bool SetPassword(char *, int, int, int, int);
		bool SetGecos(char *) { return false; }	// dummy
		bool PutEntry(FILE *);	
	
	private:
		static bool GetEntry(char *, void *);
	
	private:
		struct spwd strEntry;
};

#endif // __SHADOW_HPP__