/*
 * Copyright 1999, Alexander Feldman <alex@varna.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Alexander Feldman nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ALEXANDER FELDMAN AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL ALEXANDER FELDMAN OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "passwdc.h"
#include "globals.h"
#include "sockio.h"
#include "conf.hpp"
#include "key.hpp"
#include "rsa.hpp"

#define strzero(p) memset(p, 0, strlen(p))
#define memzero(p) memset(p, 0, sizeof(p))

int iMax = 99999;
int iMin = 0;
int iWarn = -1;
int iInact = -1;

bool fgGroup = false;

// Function prototypes
int option_value(char *, bool *, int, int, char *);

int main(int argc, char **argv)
{
	static char szServers[512];
	static char szBuf[128];
	
	static char szPassword[128];
	static char szNewPassword1[128];
	static char szNewPassword2[128];
	
	struct sockaddr_in client;
	struct hostent *pHost;
	
	char *pszLogin = (char *)NULL;
	char *pszTarget = (char *)NULL;
	char *pszServer;
	char *pszPassword;
	char *pszNewPassword1;
	char *pszNewPassword2;
	
	int sockfd;
	
	int iMaxLen;
	int iMinLen;
	
	int iServerPort;
	
	int iResult = 0;
	
	int iParams = 0;
	
	char *pszConfig = getenv("PASSWDCCONF");
	char *pszValue = NULL;
	bool fgError = false;

	if (argc > 1)
		for (int i = 1; i < argc; i++)
			if ('-' == argv[i][0]) {				// We have option
				switch (argv[i][1]) {
					case 'c':							// We have other config file
						if ('\0' != argv[i][2]) {	// The user has not typed space after -c
							pszConfig = &argv[i][2];
						} else {
							pszConfig = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
							if (NULL == pszConfig) {
								fprintf(stderr, "Invalid configuration file specified...\n");
								fgError = true;
							}
							i += 1;
						}
					break;
					case 'x':
						if ('\0' != argv[i][2]) {	// The user has not typed space after -c
							pszValue = &argv[i][2];
						} else {
							pszValue = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
							if (NULL == pszValue) {
								fprintf(stderr, "Invalid value for option: -x...\n");
								fgError = true;
							}
							i += 1;
						}
						if (NULL != pszValue)
							iMax = option_value(pszValue, &fgError, 0, 99999, "-x");
					break;
					case 'n':
						if ('\0' != argv[i][2]) {	// The user has not typed space after -c
							pszValue = &argv[i][2];
						} else {
							pszValue = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
							if (NULL == pszValue) {
								fprintf(stderr, "Invalid value for option: -n...\n");
								fgError = true;
							}
							i += 1;
						}
						if (NULL != pszValue)
							iMin = option_value(pszValue, &fgError, 0, 99999, "-n");
					break;
					case 'w':
						if ('\0' != argv[i][2]) {	// The user has not typed space after -c
							pszValue = &argv[i][2];
						} else {
							pszValue = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
							if (NULL == pszValue) {
								fprintf(stderr, "Invalid value for option: -w...\n");
								fgError = true;
							}
							i += 1;
						}
						if (NULL != pszValue)
							iWarn = option_value(pszValue, &fgError, 0, 99999, "-w");
					break;
					case 'i':
						if ('\0' != argv[i][2]) {	// The user has not typed space after -c
							pszValue = &argv[i][2];
						} else {
							pszValue = i < (argc - 1) ? argv[i + 1] : (char *)NULL;
							if (NULL == pszValue) {
								fprintf(stderr, "Invalid value for option: -i...\n");
								fgError = true;
							}
							i += 1;
						}
						if (NULL != pszValue)
							iInact = option_value(pszValue, &fgError, 0, 99999, "-i");
					break;
					case 'g':
						if ('\0' != argv[i][2])
							fgError = true;
						fgGroup = true;
					break;
					default:
						fprintf(stderr, "Invalid option: %s...\n", argv[i]);
						fgError = true;
					break;
				}
			} else {
				if (0 == iParams)
					pszTarget = argv[i];
				else if (1 == iParams)
				  pszLogin = argv[i];
				else
					fgError = true;
				iParams += 1;
			}

	if ((argc < 1) || (argc > 13) || (true == fgError) || (NULL == pszTarget)) {
		printf("Usage: passwdc <target> [<user>] [-c <config file>]\n");
		printf("       [-x <max>] [-n <min>] [-w <warn>] [-i <inact>]\n");
		printf("       passwdc <target> [<user>] -g\n");		
		printf("Where: <target>     is the user whose password you want to change;\n");
		printf("       <login>      is your login if you are master user for the given target.\n");
		printf("Options: -c <config file>\n");
		printf("                    use alternate config file instead the default\n");
		printf("                    'passwdc.ini' system wide config file;\n");
		printf("         -x <max>   sets the maximum number of days a password remains valid;\n");
		printf("         -n <min>   sets the minimum number of days before a password may be\n");
		printf("                    changed;\n");
		printf("         -w <warn>  sets the number of days before a password expiration a\n");
		printf("                    user will receive a warning;\n");
		printf("         -i <inact> sets the number of days after password expiration, the\n");
		printf("                    account will be marked inactive.\n");
		printf("         -g         change group password.\n");
		printf("Written by: <Alexander Feldman> alex@varna.net\n\n");
		return 1;
	}

	char szConfig[MAX_PATH];
	if (NULL == pszConfig) {
		GetWindowsDirectory(szConfig, sizeof(szConfig));
		strcat(szConfig, "\\"CONF_FILE);
		pszConfig = szConfig;
	}
	
	char *pszMessage = NULL;	
	CConfigurationFile *pConfigFile = new CConfigurationFile(pszConfig, &pszMessage);
	if (NULL == pConfigFile) {
		fprintf(stderr, "Error allocating memory...\n");
		return 1;
	}
	if (NULL != pszMessage) {
		fprintf(stderr, "%s\n", pszMessage);
		delete pConfigFile;
		return 1;
	}

	WSADATA WSAData;
	int iStatus;
	if (0 != (iStatus = WSAStartup(MAKEWORD(2, 2), &WSAData))) {
		printf("Socket init error: %s\n%s.\n", WSAData.szDescription, WSAData.szSystemStatus);
		return 1;
	}
	
	iServerPort = pConfigFile->GetInteger("Global", "port", SERV_TCP_PORT);
	memzero(szServers);
	strncpy(szServers, pConfigFile->GetString("Global", "targets", "localhost"), sizeof(szServers) - 1);
	
	if (NULL == pszLogin)
		pszLogin = pszTarget;
	
	pszPassword = getpass("Old password:");
	strcpy(szPassword, pszPassword);
	strzero(pszPassword);

// In fact we don't need to check for login lengths on the source machine but
// on the target but it always better to have sample restrictions for the
// min and max lengths of a string than not to have.	
	iMinLen = 5;
	iMaxLen = 8;
	printf("Enter the new password (minimum of %d, maximum of %d characters)\n", iMinLen, iMaxLen);
	printf("Please use combination of upper and lower case letters and numbers.\n");
	
	pszNewPassword1 = getpass("New password:");
	strcpy(szNewPassword1, pszNewPassword1);	
	strzero(pszNewPassword1);
	pszNewPassword2 = getpass("Re-enter new password:");
	strcpy(szNewPassword2, pszNewPassword2);
	strzero(pszNewPassword2);
	
// We can skip overflow and error checking from getpass(), because in all
// cases it returns valid pointer and the internal buffer is 128 bytes and
// null terminated.	
	
	if (0 != strcmp(szNewPassword1, szNewPassword2)) {
		printf("They don't match; try again.\n");
		iResult = 1;
		goto err_exit;
	}
	if ((int)strlen(szNewPassword1) < iMinLen || 
		 (int)strlen(szNewPassword1) > iMaxLen) {
		printf("They don't meet length criteria; try again.\n");		
		iResult = 2;
		goto err_exit;
	}
	printf("\n");

	pszServer = strtok(szServers, ";,");
	do {
		pszServer += strspn(pszServer, " \t");			
		printf("Changing password for %s on %s\n", pszTarget, pszServer);

		char szFileName[256];
		
		char *pszKeyDirectory = pConfigFile->GetString("Security", "key_directory", "/var/spool/passwdc");
		
		strncpy(szFileName, pszKeyDirectory, sizeof(szFileName));
		strncat(szFileName, "\\", sizeof(szFileName));
		strncat(szFileName, pszServer, sizeof(szFileName));
		
		if (false == RSALoadPublicKey(szFileName)) {
			printf("Requesting public key from server %s.\n", pszServer);
			get_key(pszServer, iServerPort, pszKeyDirectory);
			RSALoadPublicKey(szFileName);
		}
	
		if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET) {
			printf("Socket error: %d\n", WSAGetLastError());
			continue;
		}

		unsigned long ulHost;
		memset(&client, 0, sizeof(client));
		client.sin_family = AF_INET;
		client.sin_port = htons((unsigned short)iServerPort);
		if ((ulHost = inet_addr(pszServer)) == INADDR_NONE) {
			pHost = gethostbyname(pszServer);
			if (NULL == pHost) {
				printf("Resolving error: %d\n", WSAGetLastError());
				continue;
			}
			client.sin_addr.s_addr = *((unsigned long int *)(pHost->h_addr));
		} else
			client.sin_addr.s_addr = ulHost;

		if (connect(sockfd, (struct sockaddr *)&client, sizeof(client)) < 0) {
			printf("Connect error: %d\n", WSAGetLastError());
			close(sockfd);
			continue;
		}

		unsigned char c = '\0';
		read_byte(sockfd, &c);
		if (WELCOME != c) {
			printf("Invalid response from server.\n");
			close(sockfd);
			continue;
		} 
		if (((int)strlen(pszLogin) + 1 != write_string(sockfd, pszLogin)) ||
			 (false == write_password(sockfd, szPassword))) {
			printf("Write error.\n");
			close(sockfd);
			continue;
		} 
		read_byte(sockfd, &c);			
		switch (c) {
			case GO_AHEAD:
				write_byte(sockfd, fgGroup ? ACTION_CHANGEGROUPPASSWORD : ACTION_CHANGEPASSWORD);
				if (((int)strlen(pszTarget) + 1 != write_string(sockfd, pszTarget)) ||
					 (false == write_password(sockfd, szNewPassword1)))
					printf("Write error.\n");
				if (!fgGroup)
					if ((false == write_integer(sockfd, iMin)) ||
						 (false == write_integer(sockfd, iMax)) ||
						 (false == write_integer(sockfd, iWarn)) ||
						 (false == write_integer(sockfd, iInact)))
						printf("Write error.\n");
				read_byte(sockfd, &c);
				switch (c) {
					case OK:
						printf("Password successfully changed.\n");
					break;
					case INVALID_USER:
						printf("Invalid target specified.\n");
					break;
					case ERR:
						printf("Password still unchanged. Server returned error.\n");
					break;
					case NO_RIGHTS:
						printf("Access denied.\n");
					break;
					case WEAK_PASSWORD:
						read_string(sockfd, szBuf, sizeof(szBuf));
						printf("Weak password: %s.\n", szBuf);
					break;
					default:
						printf("Invalid response from server.\n");
				}
			break;
			case LOGIN_FAILED:
				printf("Incorrect login or password for %s on %s.\n", pszLogin, pszServer);				
			break;
			default:
				printf("Invalid response from server.\n");
		}
		close(sockfd);
	} while (NULL != (pszServer = strtok(NULL, ";,")));

err_exit:
	strzero(szPassword);
	strzero(szNewPassword1);
	strzero(szNewPassword2);

	delete pConfigFile;

	WSACleanup();
	
	return iResult;
}

int option_value(char *pszValue, bool *pfgError, int iMin, int iMax, char *pszInfo)
{
	int iResult = 0;
	
	if ((NULL == pszValue) || ('\0' == *pszValue)) {
		fprintf(stderr, "Invalid value for option: %s...\n", pszInfo);
		*pfgError = true;
		return iResult;
	}
	char *pszInvalid;
	iResult = strtol(pszValue, &pszInvalid, 10);
	if ('\0' != *pszInvalid) {
		fprintf(stderr, "Invalid value for option: %s...\n", pszInfo);
		*pfgError = true;
	}
	if (iResult < iMin || iResult > iMax) {
		fprintf(stderr, "Out of range value for option: %s...\n", pszInfo);
		*pfgError = true;
	}
	
	return iResult;
}

char *getpass(char *pszPrompt)
{
	static char szPassword[32];

	printf("%s ", pszPrompt);
	int c, i = 0;
	while (('\r' != (c = _getch())) && (i < sizeof(szPassword) - 1))
		szPassword[i++] = (char)c;
	szPassword[i] = '\0';
	putch('\n');
	
	return szPassword;
}