/*
 * Copyright 1999, Alexander Feldman <alex@varna.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Alexander Feldman nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ALEXANDER FELDMAN AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL ALEXANDER FELDMAN OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <winsock2.h>

#include "sockio.h"
#include "rsa.hpp"

int read_string(int sockfd, char *pszString, int iMaxLength)
{
	char c;
	int i, j;
	
// TODO - It'll be more faster to read bigger junks of data.
//	       We need also some additional error handling.	
	for (i = 0; i < iMaxLength - 1; i++) {
		if (1 == (j = recv(sockfd, &c, 1, 0))) {
			if ('\0' == c)
				break;
			*pszString++ = c;
		} else if (0 == j)					// EOF
			break;
		else
			return -1;							// Error
	}

	*pszString = '\0';
	return i;
}

int write_string(int sockfd, char *pszString)
{
	int iResult = send(sockfd, pszString, strlen(pszString) + 1, 0);
	
	if ((int)(strlen(pszString) + 1) != iResult)
		return -1;								// Error
	
	return iResult;
}

int read_integer(int sockfd, int *pi)
{
	return (recv(sockfd, (char *)pi, sizeof(int), 0) == sizeof(int));	
}

int write_integer(int sockfd, int i)
{
	return (send(sockfd, (char *)&i, sizeof(i), 0) == sizeof(i));	
}

int read_byte(int sockfd, unsigned char *pc)
{
	return (recv(sockfd, (char *)pc, sizeof(unsigned char), 0) == sizeof(unsigned char));
}

int write_byte(int sockfd, unsigned char c)
{
	return (send(sockfd, (char *)&c, sizeof(unsigned char), 0) == sizeof(unsigned char));
}

void *read_data(int sockfd, int *piDataLength)
{
	if (!read_integer(sockfd, piDataLength))
		return NULL;
	
	if (*piDataLength < MAX_DATA) {
		void *pvResult = calloc(*piDataLength, 1);
		if (NULL == pvResult)
			return NULL;
		if (*piDataLength != recv(sockfd, (char *)pvResult, *piDataLength, 0)) {
			free(pvResult);
			return NULL;
		}
		return pvResult;
	}
	return NULL;
}

int write_data(int sockfd, void *pvData, int iDataLength)
{
	if (!write_integer(sockfd, iDataLength))
		return -1;
	
	int iResult = send(sockfd, (const char *)pvData, iDataLength, 0);
	
	if (iResult != iDataLength)
		return -1;
	
	return iResult;
}

char *read_password(int sockfd)
{
#ifdef CRYPTED_PASSWORDS
	int iDataLength;
	
	void *pvEncryptedData = read_data(sockfd, &iDataLength);
	if (NULL == pvEncryptedData)
		return NULL;
	
	char *pszPassword = RSADecryptString(pvEncryptedData);

	free(pvEncryptedData);

	return pszPassword;
#else
	char *pszPassword = calloc(PASSWORD_BUFFER_SIZE, 1);
	if (NULL == pszPassword)
		return NULL;
	
	if (-1 == read_string(sockfd, pszPassword, PASSWORD_BUFFER_SIZE))
		return NULL;
	
	return pszPassword;
#endif // CRYPTED_PASSWORDS
}

bool write_password(int sockfd, char *pszPassword)
{
#ifdef CRYPTED_PASSWORDS
	void *pvEncryptedData = RSAEncryptString(pszPassword);
	
	write_data(sockfd, pvEncryptedData, ((unsigned int *)pvEncryptedData)[0]);
	
	free(pvEncryptedData);

	return true;
#else
	if ((int)strlen(pszPassword) + 1 != write_string(sockfd, pszPassword))
		return false;

	return true;
#endif
}